<?php
//  Arfa-CMS
//  Copyright (c) Radius17
//  Web: https://arfa-cms.ru/
//  Commercial license https://arfa-cms.ru/article/soglashenie-ob-ispolzovanii.html
//  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//  Revision: d8a6d110 (2023-09-19 02:20:16)
// 

namespace ArfaCMS\Core;

use ArfaCMS\DB\Database;
use ArfaCMS\MVC\Module;

defined('_ARFACMS_VALID') or die("Access denied");

class ComponentInstaller extends BaseObject
{
    private $_db = null;
    private $_dirsCreated = array();
    private $_filesCopied = array();
    private $_distribPath = "";
    private $_log_counter = 0;
    private $_log = array();
    private static $_instance = null;

    private function __construct($critFail)
    {
        $this->initObj();
        $this->_db = Database::getInstance();
    }

    public static function createInstance($critFail = true)
    {
        if (self::$_instance == null) {
            self::$_instance = new self($critFail);
        }
    }

    public static function getInstance($critFail = true)
    {
        self::createInstance($critFail);
        return self::$_instance;
    }

    // ====================================================================================
    // Logs functions
    // ====================================================================================

    private function _log($msg, $type)
    {
        $this->_log_counter++;
        $this->_log[$this->_log_counter]["type"] = $type;
        $this->_log[$this->_log_counter]["text"] = $msg;
    }

    private function log_message($msg)
    {
        $this->_log($msg, "message");
    }

    private function log_error($msg)
    {
        $this->_log($msg, "error");
    }

    private function log_warning($msg)
    {
        $this->_log($msg, "warning");
    }

    public function getLog()
    {
        return $this->_log;
    }

    // ====================================================================================
    // Common  install functions
    // ====================================================================================

    private function compareVersion($type, $name, $version, $operator = "<=")
    {
        // check if xml version > installed version
        $current_version = $this->getPackageVersion($name, $type);
        if (!is_null($current_version)) {
            return version_compare($current_version, $version, $operator);
        }
        return false;
    }

    private function getPackageVersion($name, $type)
    {
        $sql = "SELECT c_version FROM #__install WHERE c_type='" . $type . "' AND c_name='" . $name . "'";
        $this->_db->setQuery($sql);
        return $this->_db->loadResult();
    }

    private function createDirectory($path, $rollback = true)
    {
        if (!Files::checkFolder($path, 0)) {
            if (Files::checkFolder($path, 1)) {
                $this->log_message(Text::_("Directory created") . ": " . $path);
                if ($rollback) $this->_dirsCreated [] = $path;
                return true;
            } else {
                $this->log_error(Text::_("Unable to create directory") . ": " . $path);
                return false;
            }
        } elseif (Files::checkFolder($path, 0)) {
            return true;
        } else {
            $this->log_error(Text::_("Unable to create directory") . ": " . $path);
            return false;
        }
    }

    private function createDirectories($root, $dirs)
    {
        $success = true;
        foreach ($dirs as $dirName => $dirData) {
            $path = $root . DS . $dirName;
            if (!$this->createDirectory($path)) {
                $success = false;
                break;
            }
            $success = $this->createDirectories($path, $dirData->dirs);
            if (!$success) break;
        }
        return $success;
    }

    private function unInstallFiles($dstRoot, $dirs)
    {
        $success = true;
        foreach ($dirs as $dirName => $dirData) {
            $dstDir = $dstRoot . DS . $dirName;
            if (!$this->deleteFiles($dstDir, $dirData->files)) {
                return false;
            }
            $success = $this->unInstallFiles($dstDir, $dirData->dirs);
            if (!$success) return false;
        }
        return $success;
    }

    private function deleteFiles($dstDir, $files)
    {
        foreach ($files as $fileSrc => $fileDst) {
            $dstPath = $dstDir . DS . $fileDst;
            $this->log_message(Text::_("Deleting file ") . " " . $dstPath);
            if (!@unlink($dstPath)) return false;
        }
        return true;
    }

    private function installFiles($srcRoot, $dstRoot, $dirs)
    {
        $success = true;
        foreach ($dirs as $dirName => $dirData) {
            $srcDir = $srcRoot . DS . $dirName;
            $dstDir = $dstRoot . DS . $dirName;
            if (!$this->copyFiles($srcDir, $dstDir, $dirData->files)) {
                return false;
            }
            $success = $this->installFiles($srcDir, $dstDir, $dirData->dirs);
            if (!$success) return false;
        }
        return $success;
    }

    private function copyFiles($srcDir, $dstDir, $files)
    {
        foreach ($files as $fileSrc => $fileDst) {
            $srcPath = $srcDir . DS . $fileSrc;
            $dstPath = $dstDir . DS . $fileDst;
            if (!$this->installFile($srcPath, $dstPath)) return false;
        }
        return true;
    }

    private function installFile($srcPath, $dstPath)
    {
        if (!@copy($srcPath, $dstPath)) {
            $this->log_error(Text::_("File copy failed") . ": " . $srcPath . " => " . $dstPath);
            return false;
        } else {
            $this->_filesCopied [] = $dstPath;
            $this->log_message(Text::_("File copied") . ": " . $dstPath);

            return true;
        }
    }

    private function rollbackDirs()
    {
        $dirs = array_reverse($this->_dirsCreated);
        foreach ($dirs as $path) {
            if (@rmdir($path)) {
                $this->log_message(Text::_("Directory removed") . ": " . $path);
            } else  $this->log_error(Text::_("Failed to remove directory") . ": " . $path);
        }
    }

    private function rollbackFiles()
    {
        $files = array_reverse($this->_filesCopied);
        foreach ($files as $path) {
            if (@unlink($path)) $this->log_message(Text::_("File removed") . ": " . $path);
            else $this->log_error(Text::_("Failed to remove file") . ": " . $path);
        }
    }

    // Это наверное вообще не надо, оставил как заглушку
    private function createMenus($arr_menus, $name) { return true; }

    private function createAdminMenus($arr_menus, $module)
    {
        if (count($arr_menus)) {
            $this->log_message(Text::_("Creating admin menus"));
            foreach ($arr_menus as $key => $mnugrp) {
                $sql = "SELECT mnu_id FROM #__admin_menus WHERE mnu_parent_id=0 AND mnu_name='" . $key . "' AND mnu_module='" . $module . "' LIMIT 1";
                $this->_db->setQuery($sql);
                $parent_id = (int)$this->_db->loadResult();
                if (!$parent_id) {
                    $sql = "INSERT INTO #__admin_menus (mnu_id, mnu_parent_id, mnu_name, mnu_link, mnu_module) VALUES(0, 0, '" . $key . "', '','" . $module . "')";
                    $this->_db->setQuery($sql);
                    if ($this->_db->query()) $parent_id = $this->_db->insertid();
                }
                if ($parent_id) {
                    if (count($mnugrp)) {
                        foreach ($mnugrp as $name => $link) {
                            $sql = "SELECT COUNT(mnu_id) FROM #__admin_menus WHERE mnu_parent_id='" . $parent_id . "' AND mnu_name='" . $name . "' AND mnu_module='" . $module . "'";
                            $this->_db->setQuery($sql);
                            if ($this->_db->loadResult() == 0) {
                                $sql = "INSERT INTO #__admin_menus (mnu_id, mnu_parent_id, mnu_name, mnu_link, mnu_module) VALUES('0', '" . $parent_id . "', '" . $name . "', '" . $link . "','" . $module . "')";
                                $this->_db->setQuery($sql);
                                if (!$this->_db->query()) {
                                    $this->log_error(Text::_("Failed to add admin menus"));
                                    return false;
                                }
                            } else {
                                $sql = "UPDATE #__admin_menus SET mnu_link='" . $link . "'WHERE mnu_parent_id='" . $parent_id . "' AND mnu_name='" . $name . "' AND mnu_module='" . $module . "'";
                                $this->_db->setQuery($sql);
                                if (!$this->_db->query()) {
                                    $this->log_error(Text::_("Failed to update admin menus"));
                                    return false;
                                }
                            }
                        }
                    }
                } else {
                    $this->log_error(Text::_("Failed to add admin menus"));
                    return false;
                }
            }
        }
        return true;
    }

    // ====================================================================================
    private function checkDependencies($xml){
        if (count($xml->dependsOn)) {
            foreach ($xml->dependsOn as $dName) {
                if (!Module::isInstalled($dName)) {
                    $this->log_error(Text::_('Package needs other module to be installed first') . " " . $dName);
                    return false;
                }
            }
        }
        return true;
    }
    private function performInstall_MainFiles($xml, $folder = ""){
        if(!$folder) return false;
        if (count($xml->mainDirs) || count($xml->mainFiles)) {
            $rootDir = $folder .($xml->subname ? $xml->subname . DS : "") . $xml->name;
            $this->log_message(Text::_("Root dir") . "=" . $rootDir);
            if (!$this->createDirectory($rootDir)) return false;
            if (!$this->copyFiles($this->_distribPath . DS . $xml->mainRoot, $rootDir, $xml->mainFiles)) return false;
            if (!$this->createDirectories($rootDir, $xml->mainDirs)) return false;
            if (!$this->installFiles($this->_distribPath . DS . $xml->mainRoot, $rootDir, $xml->mainDirs)) return false;
        } else {
            $this->log_error(Text::_("Files list empty"));
            return false;
        }
        return true;
    }
    private function performInstall_Packages($xml){
        // @FIXME Всегда перезаписывается, никогда не удадляется
        if (count($xml->packagesDirs) || count($xml->packagesFiles)) {
            $dstRoot = PATH_PACKAGES . 'Custom'; // @FIXME Install in packages Custom folder
            $this->log_message(Text::_("Copying packages files to") . " " . $dstRoot);
            if (!$this->createDirectory($dstRoot)) return false;
            if (!$this->copyFiles($this->_distribPath . DS . $xml->packagesRoot, $dstRoot, $xml->packagesFiles)) return false;
            if (!$this->createDirectories($dstRoot, $xml->packagesDirs)) return false;
            if (!$this->installFiles($this->_distribPath . DS . $xml->packagesRoot, $dstRoot, $xml->packagesDirs)) return false;
        }
        return true;
    }
    private function performInstall_Assets($xml){
        // @FIXME Всегда перезаписывается, никогда не удадляется
        if (count($xml->assetsDirs) || count($xml->assetsFiles)) {
            $dstRoot = PATH_ASSETS . 'custom'; // @FIXME Install in assets custom folder
            $this->log_message(Text::_("Copying third-party files to") . " " . $dstRoot);
            if (!$this->createDirectory($dstRoot)) return false;
            if (!$this->copyFiles($this->_distribPath . DS . $xml->assetsRoot, $dstRoot, $xml->assetsFiles)) return false;
            if (!$this->createDirectories($dstRoot, $xml->assetsDirs)) return false;
            if (!$this->installFiles($this->_distribPath . DS . $xml->assetsRoot, $dstRoot, $xml->assetsDirs)) return false;
        }
        return true;
    }
    private function performInstall_Lang($xml, $subfolder="", $branch = "common"){
        // if(!$subfolder) return false;
        if(!$branch || (!$subfolder && !$xml->subname)) return false;
        $this->log_message(Text::_("Installing language files"));
        foreach ($xml->language as $languageName => $language_file) {
            $langSrcPath = $this->_distribPath . DS . $language_file;
            // $langDstPath = PATH_LANGUAGE . $branch . DS . $languageName . DS . $subfolder . DS . ($xml->subname ? $xml->subname . DS : "");
            $langDstPath = PATH_LANGUAGE . $branch . DS . $languageName . DS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname . DS : "");
            $langDstFilepath = $langDstPath . $xml->name . ".ini";
            if (file_exists($langSrcPath)) {
                if (!Files::checkFolder($langDstPath, true) || !$this->installFile($langSrcPath, $langDstFilepath)) {
                    return false;
                }
            }
        }
        return true;
    }
    private function performInstall_CSS($xml, $subfolder=""){
        // if(!$subfolder) return false;
        if(!$subfolder && !$xml->subname) return false;
        if (count($xml->cssDirs) || count($xml->cssFiles)) {
            // $dstRoot = PATH_CSS . $subfolder . ($xml->subname ? DS . $xml->subname : "");
            $dstRoot = PATH_CSS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname . DS : "");
            $this->log_message(Text::_("Copying css files to") . " " . $dstRoot);
            if (!$this->createDirectory($dstRoot, false)) return false;
            if (!$this->copyFiles($this->_distribPath . DS . $xml->cssRoot, $dstRoot, $xml->cssFiles)) return false;
            if (!$this->createDirectories($dstRoot, $xml->cssDirs)) return false;
            if (!$this->installFiles($this->_distribPath . DS . $xml->cssRoot, $dstRoot, $xml->cssDirs)) return false;
        }
        return true;
    }
    private function performInstall_Javascript($xml, $subfolder=""){
        // if(!$subfolder) return false;
        if(!$subfolder && !$xml->subname) return false;
        if (count($xml->jsDirs) || count($xml->jsFiles)) {
            // $dstRoot = PATH_JS . $subfolder . ($xml->subname ? DS . $xml->subname : "");
            $dstRoot = PATH_JS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname .DS : "");
            $this->log_message(Text::_("Copying javascript files to") . " " . $dstRoot);
            if (!$this->createDirectory($dstRoot, false)) return false;
            if (!$this->copyFiles($this->_distribPath . DS . $xml->jsRoot, $dstRoot, $xml->jsFiles)) return false;
            if (!$this->createDirectories($dstRoot, $xml->jsDirs)) return false;
            if (!$this->installFiles($this->_distribPath . DS . $xml->jsRoot, $dstRoot, $xml->jsDirs)) return false;
        }
        return true;
    }
    private function performInstall_Queries($xml){
        if (count($xml->queries)) {
            $update_result = true; // Only for update installation
            $this->log_message(Text::_("Applying install queries to database"));
            foreach ($xml->queries as $query) {
                if ($query->max_version == "" || $this->compareVersion($xml->type, ($xml->subname ? $xml->subname . "." : "") . $xml->name, $query->max_version, "<")) {
                    $this->_db->setQuery($query->sql);
                    if (!$this->_db->query()) {
                        if ($xml->update) {
                            $this->log_error("SQL Error :" . $this->_db->getLastError() . "<br />" . $this->_db->getQuery());
                            $update_result = false;
                        } else {
                            $this->log_error(Text::_("Failed to apply queries to database"));
                            return false;
                        }
                    }
                }
            }
            if ($xml->update && !$update_result) $this->log_error(Text::_("Failed to apply queries to database"));
        }
        return true;
    }
    private function tableInstall_Insert($xml){
        $installdata = $xml->xmlData();
        $query = "INSERT INTO #__install (c_id, c_type, c_name,	c_version, c_description, c_author, c_email, c_site, c_license, c_data)
                VALUES(NULL,'".$xml->type."','" . ($xml->subname ? $xml->subname . "." : "") . $xml->name . "',	'" . $xml->version . "', '" . $xml->description . "', '" . $xml->author . "', '" . $xml->email . "', '" . $xml->site . "', '" . addslashes($xml->license) . "', '" . addslashes($installdata) . "')";
        $this->_db->setQuery($query);
        if (!$this->_db->query()) {
            $this->log_error(Text::_("Failed to write install info into database"));
            return false;
        }
        return true;
    }
    private function tableInstall_Update($xml){
        $installdata = $xml->xmlData();
        $query = "UPDATE #__install SET c_version='" . $xml->version . "', c_description='" . $xml->description . "', c_author='" . $xml->author . "', c_email='" . $xml->email . "', c_site='" . $xml->site . "', c_license='" . addslashes($xml->license) . "', c_data='" . addslashes($installdata) . "'";
        $query .= " WHERE c_type='".$xml->type."' AND c_name='" . ($xml->subname ? $xml->subname . "." : "") . $xml->name . "'";
        $this->_db->setQuery($query);
        if (!$this->_db->query()) {
            $this->log_error(Text::_("Failed to write update info into database"));
            $this->log_error("SQL Error :" . $this->_db->getLastError() . "<br />" . $this->_db->getQuery());
            return false;
        }
        return true;
    }
    // ====================================================================================
    private function performUninstall_MainFiles($xml, $folder=""){
        if(!$folder) return false;
        if (count($xml->mainDirs) || count($xml->mainFiles)) {
            $rootDir = $folder .($xml->subname ? $xml->subname . DS : "") . $xml->name;
            $this->log_message(Text::_("Root dir") . "=" . $rootDir);
            $result = Files::removeFolder($rootDir, 1);
            if ($result) $this->log_message(Text::_("Root folder removed"));
            else {
                $this->log_error(Text::_("Failed removing root folder"));
                return false;
            }
        }
        return true;
    }
    private function performUninstall_RemoveLang($xml, $subfolder="", $branch = "common"){
        // if(!$subfolder) return false;
        if(!$branch || (!$subfolder && !$xml->subname)) return false;
        foreach ($xml->language as $languageName => $language) {
            // $langDstPath = PATH_LANGUAGE . $branch . DS . $languageName . DS . $subfolder . DS . ($xml->subname ? $xml->subname . DS : "") . $xml->name . ".ini";
            $langDstPath = PATH_LANGUAGE . $branch . DS . $languageName . DS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname . DS : "") . $xml->name . ".ini";
            if (file_exists($langDstPath)) {
                $this->log_message(Text::_("Removing language file") . ": " . $langDstPath);
                if (!unlink($langDstPath)) {
                    $this->log_error(Text::_("Failed removing language file"));
                    return false;
                }
            }
        }
        return true;
    }
    private function performUninstall_RemoveCSS($xml, $subfolder=""){
        // if(!$subfolder) return false;
        if(!$subfolder && !$xml->subname) return false;
        if (count($xml->cssFiles)) {
            foreach ($xml->cssFiles as $file) {
                // $cssPath = PATH_CSS . $subfolder . DS . ($xml->subname ? $xml->subname.DS : "") . $file;
                $cssPath = PATH_CSS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname . DS : "") . $file;
                if (file_exists($cssPath)) {
                    $this->log_message(Text::_("Removing css file") . ": " . $cssPath);
                    if (!unlink($cssPath)) {
                        $this->log_error(Text::_("Failed removing css file"));
                        return false;
                    }
                } else $this->log_error(Text::_("File absent") . ": " . $cssPath);
            }
        }
        return true;
    }
    private function performUninstall_RemoveJavascript($xml, $subfolder=""){
        // if(!$subfolder) return false;
        if(!$subfolder && !$xml->subname) return false;
        if (count($xml->jsFiles)) {
            foreach ($xml->jsFiles as $file) {
                // $jsPath = PATH_JS . $subfolder . DS . ($xml->subname ? $xml->subname.DS : "") . $file;
                $jsPath = PATH_JS . ($subfolder ? $subfolder . DS : "") . ($xml->subname ? $xml->subname . DS : "") . $file;
                if (file_exists($jsPath)) {
                    $this->log_message(Text::_("Removing javascript file") . ": " . $jsPath);
                    if (!unlink($jsPath)) {
                        $this->log_error(Text::_("Failed removing javascript file"));
                        return false;
                    }
                } else $this->log_error(Text::_("File absent") . ": " . $jsPath);
            }
        }
        return true;
    }
    private function performUninstall_Queries($xml){
        if (count($xml->unqueries)) {
            $this->log_message(Text::_("Applying uninstall queries to database"));
            foreach ($xml->unqueries as $sql) {
                $this->_db->setQuery($sql);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to apply uninstall queries to database"));
                    return false;
                }
            }
        }
        return true;
    }
    private function tableInstall_Delete($psid){
        $query = "DELETE FROM #__install WHERE c_id=" . $psid;
        $this->_db->setQuery($query);
        if ($this->_db->query()) $this->log_message(Text::_("Cleaned data from install table"));
        else $this->log_error(Text::_("Failed to clean data from install table"));
    }
    // ====================================================================================
    // START INSTALL HERE
    // ====================================================================================
    public function install($distribDir)
    {
        $success = false;
        $xmlPath = $distribDir . DS . "install.xml";
        $xml = new InstallerXML($xmlPath);
        if ($xml->isLoaded()) {
            $this->_distribPath = $distribDir;
            switch ($xml->type) {
                case "module":
                    $is_installed = Module::isInstalled($xml->name);
                    $success = $this->installModule($xml, $is_installed);
                    break;
                case "widget":
                    $is_installed = Widget::isInstalled($xml->name);
                    $success = $this->installWidget($xml, $is_installed);
                    break;
                case "plugin":
                    $is_installed = Plugin::isInstalled($xml->subname, $xml->name);
                    $success = $this->installPlugin($xml, $is_installed);
                    break;
                case "package":
                    $success = $this->installPackage($xml);
                    break;
                case "payment":
                    $is_installed = orderPayment::isInstalled($xml->name);
                    $success = $this->installPayment($xml, $is_installed);
                    break;
                case "delivery":
                    $is_installed = orderDelivery::isInstalled($xml->name);
                    $success = $this->installDelivery($xml, $is_installed);
                    break;
                case "template":
                    $success = $this->installTemplate($xml);
                    break;
                default:
                    break;
            }
            if ($success) return true;
            else {
                if (!$xml->update || !$is_installed) {
                    $this->rollbackFiles();
                    $this->rollbackDirs();
                }
                return false;
            }
        } else {
            $this->log_error(Text::_("Can not load") . " " . $xmlPath);
            return false;
        }
    }
    public function uninstall($xmlFile, $psid)
    {
        $success = false;
        $xml = new InstallerXML($xmlFile);
        if ($xml->isLoaded()) {
            switch ($xml->type) {
                case "module":
                    $success = $this->uninstallModule($xml, $psid);
                    break;
                case "widget":
                    $success = $this->uninstallWidget($xml, $psid);
                    break;
                case "plugin":
                    $success = $this->uninstallPlugin($xml, $psid);
                    break;
                case "package":
                    $success = $this->uninstallPackage($xml, $psid);
                    break;
                case "payment":
                    $success = $this->uninstallPayment($xml, $psid);
                    break;
                case "delivery":
                    $success = $this->uninstallDelivery($xml, $psid);
                    break;
                case "template":
                    $success = $this->uninstallTemplate($xml, $psid);
                    break;
                default:
                    break;
            }
            return $success;
        } else {
            $this->log_error(Text::_("Can not load") . " " . $xmlFile);
            return false;
        }
    }

    // ====================================================================================

    private function installModule(&$xml, $is_installed)
    {
        if (!$xml->name) {
            $this->log_message(Text::_("Installing module") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing module") . " " . $xml->nameAsIs);
        }
        // ------------------------------------------------------
        if ($xml->name == "system" || ($is_installed && !$xml->update)) {
            $this->log_error(Text::_('Module is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("module", $xml->name, $xml->version)) {
            $this->log_error(Text::_('Module version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            if(!$this->performInstall_MainFiles($xml, PATH_MODULES)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if(!$this->performInstall_Lang($xml, "modules")) return false;
            if(!$this->performInstall_CSS($xml, "modules")) return false;
            if(!$this->performInstall_Javascript($xml, "modules")) return false;
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            // Build config
            $moduleConfig = "";
            foreach ($xml->params as $cfgKey => $cfgValue) {
                $moduleConfig .= "$cfgKey=$cfgValue;";
            }
            $moduleConfig = substr($moduleConfig, 0, strlen($moduleConfig) - 1);
            if (!$is_installed) {
                $query = "INSERT INTO #__modules (m_id, m_name, m_show_breadcrumb, m_config, m_incl_map, m_deleted) VALUES (NULL, '" . $xml->name . "', 1, '" . $moduleConfig . "', 0, 0)";
                $this->_db->setQuery($query);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to install module into database"));
                    return false;
                }
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            // ------------------------------------------------------
            // Module access ACL
            ACLObject::createACLObject($xml->name, $xml->name . "Module", "Module access", 0);
            ACLObject::createACLObject($xml->name, $xml->name . "Module", "Module access", 1);
            // Developer-defined ACL
            foreach ($xml->frontACL as $aclName => $aclDescr) {
                ACLObject::createACLObject($xml->name, $aclName, $aclDescr, 0);
            }
            foreach ($xml->adminACL as $aclName => $aclDescr) {
                ACLObject::createACLObject($xml->name, $aclName, $aclDescr, 1);
            }
            // ------------------------------------------------------
            // Finally create menus
            $this->createAdminMenus($xml->adminMenus, $xml->name);
            $this->createMenus($xml->frontMenus, $xml->name);
            // ------------------------------------------------------
            return true;
        }
    }

    private function installPlugin(&$xml, $is_installed)
    {
        if (!$xml->name || !$xml->subname) {
            $this->log_message(Text::_("Installing plugin") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing plugin") . " " . $xml->subname . "." . $xml->name);
        }
        // ------------------------------------------------------
        if (Plugin::isCorrupted($xml->subname, $xml->name)) {
            $this->log_error(Text::_('Plugin installation corrupted'));
            return false;
        } elseif ($is_installed && !$xml->update) {
            $this->log_error(Text::_('Plugin is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("plugin", $xml->subname . "." . $xml->name, $xml->version)) {
            $this->log_error(Text::_('Plugin version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            if(!$this->performInstall_MainFiles($xml, PATH_PLUGINS)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if(!$this->performInstall_Lang($xml, "plugins")) return false;
            if(!$this->performInstall_CSS($xml, "plugins")) return false;
            if(!$this->performInstall_Javascript($xml, "plugins")) return false;
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            if (!$is_installed) {
                $query = "INSERT INTO #__plugins (p_id, p_path, p_name, p_params, p_ordering, p_enabled, p_deleted) VALUES (NULL, '" . $xml->subname . "', '" . $xml->name . "', '', 0, 0, 0)";
                $this->_db->setQuery($query);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to install plugin into database"));
                    return false;
                }
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            return true;
        }
    }

    private function installWidget(&$xml, $is_installed)
    {
        if (!$xml->name) {
            $this->log_message(Text::_("Installing widget") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing widget") . " " . $xml->nameAsIs);
        }
        // ------------------------------------------------------
        if ($is_installed && !$xml->update) {
            $this->log_error(Text::_('Widget is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("widget", $xml->name, $xml->version)) {
            $this->log_error(Text::_('Widget version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            if(!$this->performInstall_MainFiles($xml, PATH_WIDGETS)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if(!$this->performInstall_Lang($xml, "widgets")) return false;
            if(!$this->performInstall_CSS($xml, "widgets")) return false;
            if(!$this->performInstall_Javascript($xml, "widgets")) return false;
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            if (!$is_installed) {
                $query = "INSERT INTO #__widgets (w_id, w_name, w_side) VALUES (NULL,'" . $xml->name . "', " . $xml->wside . ")";
                $this->_db->setQuery($query);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to install widget into database"));
                    return false;
                }
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            return true;
        }
    }

    private function installPayment(&$xml, $is_installed)
    {
        if (!$xml->name) {
            $this->log_message(Text::_("Installing payment type") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing payment type") . " " . $xml->nameAsIs);
        }
        // ------------------------------------------------------
        if ($is_installed && !$xml->update) {
            $this->log_error(Text::_('Payment type is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("payment", $xml->name, $xml->version)) {
            $this->log_error(Text::_('Payment type version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            $dstRoot = PATH_PACKAGES . 'ArfaCMS' . DS . 'Payment';
            if (!$this->copyFiles($this->_distribPath . DS . $xml->mainRoot, $dstRoot, $xml->mainFiles)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if(!$this->performInstall_Lang($xml, "payments")) return false;
            if(!$this->performInstall_CSS($xml, "payments")) return false;
            if(!$this->performInstall_Javascript($xml, "payments")) return false;
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            if (!$is_installed) {
                $query = "INSERT INTO #__pts (p_name) VALUES ('" . $xml->nameAsIs . "')";
                $this->_db->setQuery($query);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to install payment type into database"));
                    return false;
                }
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            return true;
        }
    }

    private function installDelivery(&$xml, $is_installed)
    {
        if (!$xml->name) {
            $this->log_message(Text::_("Installing delivery type") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing delivery type") . " " . $xml->nameAsIs);
        }
        // ------------------------------------------------------
        if ($is_installed && !$xml->update) {
            $this->log_error(Text::_('Delivery type is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("delivery", $xml->name, $xml->version)) {
            $this->log_error(Text::_('Delivery type version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            $dstRoot = PATH_PACKAGES . 'ArfaCMS' . DS . 'Delivery';
            if (!$this->copyFiles($this->_distribPath . DS . $xml->mainRoot, $dstRoot, $xml->mainFiles)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if(!$this->performInstall_Lang($xml, "deliveries")) return false;
            if(!$this->performInstall_CSS($xml, "deliveries")) return false;
            if(!$this->performInstall_Javascript($xml, "deliveries")) return false;
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            if (!$is_installed) {
                $query = "INSERT INTO #__dts (d_name) VALUES ('" . $xml->nameAsIs . "')";
                $this->_db->setQuery($query);
                if (!$this->_db->query()) {
                    $this->log_error(Text::_("Failed to install delivery type into database"));
                    return false;
                }
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            return true;
        }
    }

    private function installPackage(&$xml, $subfolder="")
    {
        // @FIXME Проверить установку пакета
        $is_installed = false; // @FIXME Всегда новый ???
        // ------------------------------------------------------
        if (!$xml->name) {
            $this->log_message(Text::_("Installing package") . "...");
            $this->log_error(Text::_('Wrong name or path'));
            return false;
        } else {
            $this->log_message(Text::_("Installing package") . " " . $xml->nameAsIs);
        }
        // ------------------------------------------------------
        if ($is_installed && !$xml->update) {
            $this->log_error(Text::_('Package is allready installed'));
            return false;
        } elseif ($is_installed && $xml->update && !$this->compareVersion("package", $xml->name, $xml->version)) {
            $this->log_error(Text::_('Package version is less then allready installed'));
            return false;
        } else {
            // ------------------------------------------------------
            if(!$this->checkDependencies($xml)) return false;
            if(!$this->performInstall_Packages($xml)) return false;
            if(!$this->performInstall_Assets($xml)) return false;
            if($subfolder) {
                if (!$this->performInstall_Lang($xml, $subfolder)) return false;
                if (!$this->performInstall_CSS($xml, $subfolder)) return false;
                if (!$this->performInstall_Javascript($xml, $subfolder)) return false;
            }
            if(!$this->performInstall_Queries($xml)) return false;
            // ------------------------------------------------------
            if (!$is_installed) {
                if(!$this->tableInstall_Insert($xml)) return false;
            } elseif ($xml->update) {
                if(!$this->tableInstall_Update($xml)) return false;
            }
            return true;
        }
    }

    public function installTemplate(&$xml){
        // Переводы ставим через performInstall_MainFiles
        // @FIXME Еще не сделано
        return false;
    }
    // ====================================================================================
    public function uninstallModule(&$xml, $psid)
    {
        if (!$xml->name) return false;
        $this->log_message(Text::_("Uninstalling module") . " " . $xml->nameAsIs);
        // ------------------------------------------------------
        if(!$this->performUninstall_MainFiles($xml, PATH_MODULES)) return false;
        if(!$this->performUninstall_RemoveLang($xml, "modules")) return false;
        if(!$this->performUninstall_RemoveCSS($xml, "modules")) return false;
        if(!$this->performUninstall_RemoveJavascript($xml, "modules")) return false;
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        $query = "DELETE FROM #__modules WHERE m_name='" . $xml->name . "'" . $this->_db->getDelimiter();
        $query.= "DELETE FROM #__admin_menus WHERE mnu_module='" . $xml->name . "'" . $this->_db->getDelimiter();
        $query.= "DELETE FROM #__acl_objects WHERE ao_module_name='" . $xml->name . "'" . $this->_db->getDelimiter();
        $this->_db->setQuery($query);
        if($this->_db->query_batch(false, true))  $this->log_message(Text::_("Cleaned data from database"));
        else $this->log_error(Text::_("Failed to clean data from database"));
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallPlugin(&$xml, $psid)
    {
        if (!$xml->name || !$xml->subname) return false;
        $this->log_message(Text::_("Uninstalling plugin") . " " . $xml->subname . "." . $xml->name);
        // ------------------------------------------------------
        if(!$this->performUninstall_MainFiles($xml, PATH_PLUGINS)) return false;
        if(!$this->performUninstall_RemoveLang($xml, "plugins")) return false;
        if(!$this->performUninstall_RemoveCSS($xml, "plugins")) return false;
        if(!$this->performUninstall_RemoveJavascript($xml, "plugins")) return false;
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        // Main table
        $query = "DELETE FROM #__plugins WHERE p_name='" . $xml->name . "' AND p_path='" . $xml->subname . "'" . $this->_db->getDelimiter();
        $this->_db->setQuery($query);
        if($this->_db->query_batch(false, true))  $this->log_message(Text::_("Cleaned data from database"));
        else $this->log_error(Text::_("Failed to clean data from database"));
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallWidget(&$xml, $psid)
    {
        if (!$xml->name) return false;
        $this->log_message(Text::_("Uninstalling widget") . " " . $xml->nameAsIs);
        // ------------------------------------------------------
        if(!$this->performUninstall_MainFiles($xml, PATH_WIDGETS)) return false;
        if(!$this->performUninstall_RemoveLang($xml, "widgets")) return false;
        if(!$this->performUninstall_RemoveCSS($xml, "widgets")) return false;
        if(!$this->performUninstall_RemoveJavascript($xml, "widgets")) return false;
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        // Main tables
        $query = "DELETE FROM #__widgets WHERE w_name='" . $xml->name . "'" . $this->_db->getDelimiter();
        $query.= "DELETE FROM #__widgets_active WHERE aw_name='" . $xml->name . "'" . $this->_db->getDelimiter();
        $this->_db->setQuery($query);
        if($this->_db->query_batch(false, true))  $this->log_message(Text::_("Cleaned data from database"));
        else $this->log_error(Text::_("Failed to clean data from database"));
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallDelivery(&$xml, $psid)
    {
        if (!$xml->name) return false;
        $this->log_message(Text::_("Uninstalling delivery type") . " " . $xml->nameAsIs);
        // ------------------------------------------------------
        $dstRoot = PATH_PACKAGES . 'ArfaCMS' . DS . 'Delivery';
        if (!$this->deleteFiles($dstRoot, $xml->mainFiles)) return false;
        if(!$this->performUninstall_RemoveLang($xml, "deliveries")) return false;
        if(!$this->performUninstall_RemoveCSS($xml, "deliveries")) return false;
        if(!$this->performUninstall_RemoveJavascript($xml, "deliveries")) return false;
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        // Main tables
        $query = "DELETE FROM #__dts WHERE LOWER(d_name)='" . strtolower($xml->name) . "'" . $this->_db->getDelimiter();
        $query.= "DELETE FROM #__dts_active WHERE LOWER(dt_name)='" . strtolower($xml->name) . "'" . $this->_db->getDelimiter();
        $this->_db->setQuery($query);
        if($this->_db->query_batch(false, true))  $this->log_message(Text::_("Cleaned data from database"));
        else $this->log_error(Text::_("Failed to clean data from database"));
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallPayment(&$xml, $psid)
    {
        if (!$xml->name) return false;
        $this->log_message(Text::_("Uninstalling payment type") . " " . $xml->nameAsIs);
        // ------------------------------------------------------
        $dstRoot = PATH_PACKAGES . 'ArfaCMS' . DS . 'Payment';
        if (!$this->deleteFiles($dstRoot, $xml->mainFiles)) return false;
        if(!$this->performUninstall_RemoveLang($xml, "payments")) return false;
        if(!$this->performUninstall_RemoveCSS($xml, "payments")) return false;
        if(!$this->performUninstall_RemoveJavascript($xml, "payments")) return false;
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        // Main tables
        $query = "DELETE FROM #__pts WHERE LOWER(p_name)='" . strtolower($xml->name) . "'" . $this->_db->getDelimiter();
        $query.= "DELETE FROM #__pts_active WHERE LOWER(pt_name)='" . strtolower($xml->name) . "'" . $this->_db->getDelimiter();
        $this->_db->setQuery($query);
        if($this->_db->query_batch(false, true))  $this->log_message(Text::_("Cleaned data from database"));
        else $this->log_error(Text::_("Failed to clean data from database"));
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallPackage(&$xml, $psid, $subfolder="")
    {
        // @FIXME Проверить удаление пакета
        if (!$xml->name) return false;
        $this->log_message(Text::_("Uninstalling package") . " " . $xml->nameAsIs);
        // ------------------------------------------------------
        if($subfolder) {
            if (!$this->performUninstall_RemoveLang($xml, $subfolder)) return false;
            if (!$this->performUninstall_RemoveCSS($xml, $subfolder)) return false;
            if (!$this->performUninstall_RemoveJavascript($xml, $subfolder)) return false;
        }
        if(!$this->performUninstall_Queries($xml)) return false;
        // ------------------------------------------------------
        // Cleaning tables
        // ------------------------------------------------------
        $this->tableInstall_Delete($psid);
        return true;
    }
    public function uninstallTemplate(&$xml, $psid){
        // @FIXME Еще не сделано
        return false;
    }
}
?>