<?php
//  Arfa-CMS
//  Copyright (c) Radius17
//  Web: https://arfa-cms.ru/
//  Commercial license https://arfa-cms.ru/article/soglashenie-ob-ispolzovanii.html
//  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//  Revision: 0f2b12bf (2024-10-19 11:22:48)
// 

namespace ArfaCMS\Core;

defined('_ARFACMS_VALID') or die("Access denied");

class Request
{
    private static function getFrom($arr, $key, $defVal = '', $sef = false)
    {
        if ($sef || \seoConfig::$sefMode) {
            $rval = Router::getInstance()->getRequest($key);
            if (!is_null($rval)) return $rval;
        }
        return Util::getArrParam($arr, $key, $defVal);
    }


    public static function getRealValue($key, $defVal = '', $type = 'request')
    {
        return self::get($key, $defVal, $type);
    }

    private static function get($key, $defVal, $type)
    {
        switch (strtolower($type)) {
            case 'get':
                return self::getFrom($_GET, $key, $defVal, true);
                break;
            case 'post':
                return self::getFrom($_POST, $key, $defVal, true);
                break;
            case 'session':
                return self::getFrom($_SESSION, $key, $defVal);
                break;
            case 'cookie':
                return self::getFrom($_COOKIE, $key, $defVal);
                break;
            case 'server':
                return self::getFrom($_SERVER, $key, $defVal);
                break;
            case 'request':
            default:
                return self::getFrom($_REQUEST, $key, $defVal, true);
                break;
        }
    }

    public static function checkExists($key, $type = 'request')
    {
        switch (strtolower($type)) {
            case 'get':
                return array_key_exists($key, $_GET);
                break;
            case 'post':
                return array_key_exists($key, $_POST);
                break;
            case 'session':
                return array_key_exists($key, $_SESSION);
                break;
            case 'cookie':
                return array_key_exists($key, $_COOKIE);
                break;
            case 'server':
                return array_key_exists($key, $_SERVER);
                break;
            case 'request':
            default:
                return array_key_exists($key, $_REQUEST);
                break;
        }
    }

    /**
     * @param $source
     * @param $cast "safe"|"bool"|"int"|"bigint"|"float"
     * @return array|int|string
     */
    public static function makeCast($source, $cast = "safe")
    {
        if (is_array($source)) {
            if (count($source)) {
                foreach ($source as $key => $val) {
                    $source[$key] = self::makeCast($val, $cast);
                }
            }
            return $source;
        } else {
            switch ($cast) {
                case "bool":
                    return Util::toBool($source);
                    break;
                case "int":
                    return intval(str_replace(array(' '), array(''), $source));
                    break;
                case "bigint":
                    return Util::toBigInt(str_replace(array(' '), array(''), $source));
                    break;
                case "float":
                    return floatval(str_replace(array(' ', ','), array('', '.'), $source));
                    break;
                case "safe":
                default:
                    return htmlspecialchars($source, ENT_QUOTES, 'UTF-8');
                    break;
            }
        }
    }

    public static function getBigInt($key, $defVal = 0, $type = 'request')
    {
        return Util::toBigInt(str_replace(array(' '), array(''), self::get($key, intval($defVal), $type)));
        // return self::makeCast(self::get($key, $defVal, $type), "bigint"); // Move this to separate function, if you need array
    }

    public static function getInt($key, $defVal = 0, $type = 'request')
    {
        return intval(str_replace(array(' '), array(''), self::get($key, intval($defVal), $type)));
        // return self::makeCast(self::get($key, $defVal, $type), "int"); // Move this to separate function, if you need array
    }

    public static function getFloat($key, $defVal = 0, $type = 'request')
    {
        return floatval(str_replace(array(' ', ','), array('', '.'), self::get($key, $defVal, $type)));
        // return self::makeCast(self::get($key, $defVal, $type), "float"); // Move this to separate function, if you need array
    }

    public static function getBool($key, $defVal = false, $type = 'request')
    {
        return Util::toBool(self::get($key, $defVal, $type));
        // return self::makeCast(self::get($key, $defVal, $type), "bool"); // Move this to separate function, if you need array
    }

    public static function getDateTime($key, $defVal = false, $type = 'request')
    {
        $result = self::getSafe($key, "", $type);
        if ($result) return Date::toSQL($result);
        else return $defVal;
    }

    public static function getSafe($key, $defVal = '', $type = 'request')
    {
        return self::makeCast(self::get($key, $defVal, $type));
    }

    public static function getDate($key, $defVal = false, $type = 'request')
    {
        $result = self::getSafe($key, "", $type);
        if ($result) return Date::toSQL($result, true);
        else return $defVal;
    }

    public static function getMethod()
    {
        return strtoupper(self::getSafe("REQUEST_METHOD", "", "server"));
    }

    public static function getHTML($key, $defVal = '', $type = 'request')
    {
        return addslashes(self::get($key, $defVal, $type));
    }

    // Form tokens
    public static function getVerificationToken($makeHash = true){
        $requestVerificationSalt = Util::generateRandomString(8, "", 2);
        $requestVerificationToken = Session::getVar("__requestVerificationToken");
        if(is_null($requestVerificationToken)){
            $requestVerificationToken = self::setVerificationToken();
        }
        return $makeHash ? $requestVerificationSalt.":".md5($requestVerificationSalt.":".$requestVerificationToken) : $requestVerificationToken;
    }

    public static function generateVerificationToken(){
        return Util::generateRandomString(32, "", 2);
    }

    public static function checkVerificationToken($token = "", $allowPostOnly = true){
        if (!$token) $token = Request::getSafe("_requestVerificationToken"); // @TODO May be not ???
        if (!$token) return false;
        if ($allowPostOnly && Request::getMethod() !== "POST") return false;

        $tokenParts = explode(":", $token);
        if (count($tokenParts) != 2) return false;

        $requestVerificationSalt = trim($tokenParts[0]);
        $requestVerificationHash = trim($tokenParts[1]);
        if (!$requestVerificationSalt || !$requestVerificationHash) return false;

        $requestVerificationToken = Session::getVar("__requestVerificationToken");
        if (!$requestVerificationToken) return false;

        return $requestVerificationHash == md5($requestVerificationSalt.":".$requestVerificationToken);
    }

    public static function setVerificationToken(){
        $requestVerificationToken = self::generateVerificationToken();
        Session::setVar("__requestVerificationToken", $requestVerificationToken);
        return $requestVerificationToken;
    }
}
?>