<?php
//  Arfa-CMS
//  Copyright (c) Radius17
//  Web: https://arfa-cms.ru/
//  Commercial license https://arfa-cms.ru/article/soglashenie-ob-ispolzovanii.html
//  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//  Revision: d832b2ab (2024-10-20 09:51:18)
// 

namespace ArfaCMS\MVC;

use ArfaCMS\Core\ACLObject;
use ArfaCMS\Core\BaseObject;
use ArfaCMS\Core\Debugger;
use ArfaCMS\Core\FieldSet;
use ArfaCMS\Core\Files;
use ArfaCMS\Core\Params;
use ArfaCMS\Core\Portal;
use ArfaCMS\Core\Registry;
use ArfaCMS\Core\Request;
use ArfaCMS\Core\Router;
use ArfaCMS\Core\Text;
use ArfaCMS\Core\User;
use ArfaCMS\Core\Util;
use ArfaCMS\DB\Database;

defined('_ARFACMS_VALID') or die("Access denied");

class Module extends BaseObject
{
    //--------------- Factory implementation --------------------
    private static $_loadedModules = array();
    private static $_loadMode = 'main';
    private static $coreModules = array('aclmgr', 'conf', 'user');
    protected $_paramsMask = array(); // vtype, vdefault, fill_default, source, descr

    public function addParam($name, $vtype, $vdefault, $fill_default = false, $source = null, $description = null)
    {
        $this->_paramsMask[$name]["vtype"] = $vtype;
        $this->_paramsMask[$name]["vdefault"] = $vdefault;
        $this->_paramsMask[$name]["fill_default"] = $fill_default;
        if (!is_null($description)) $this->_paramsMask[$name]["descr"] = $description;
        if (!is_null($source)) $this->_paramsMask[$name]["source"] = $source;
    }

    protected function setParamsMask()
    {
        $controllers_front_arr = array();
        $controllers_front = Files::getFiles(PATH_MODULES . $this->getName() . DS . "controllers", array(), false);
        if (is_array($controllers_front)) {
            $controllers_front_keys = array_keys($controllers_front);
            $controllers_front_keys = str_replace('.php', '', $controllers_front_keys);
            $controllers_front_arr = array_combine($controllers_front_keys, $controllers_front_keys);
        }
        $views_front_arr = array();
        $views_front = Files::getFolders(PATH_MODULES . $this->getName() . DS . "views" . DS . "layouts", false, false);
        if (is_array($views_front)) {
            $views_front_keys = array_keys($views_front);
            $views_front_arr = array_combine($views_front_keys, $views_front_keys);
        }
        $controllers_admin_arr = array();
        $controllers_admin = Files::getFiles(PATH_MODULES . $this->getName() . DS . "controllers" . DS . $this->_adm_folder, array(), false);
        if (is_array($controllers_admin)) {
            $controllers_admin_keys = array_keys($controllers_admin);
            $controllers_admin_keys = str_replace('.php', '', $controllers_admin_keys);
            $controllers_admin_arr = array_combine($controllers_admin_keys, $controllers_admin_keys);
        }
        $views_admin_arr = array();
        $views_admin = Files::getFolders(PATH_MODULES . $this->getName() . DS . "views" . DS . $this->_adm_folder . "layouts", false, false);
        if (is_array($views_admin)) {
            $views_admin_keys = array_keys($views_admin);
            $views_admin_arr = array_combine($views_admin_keys, $views_admin_keys);
        }
        $this->addParam("Page_size", "integer", \siteConfig::$recordsPerPage, true);
        $this->addParam("Default_controller", "select", reset($controllers_front_arr), true, $controllers_front_arr);
        $this->addParam("Default_view", "select", "", true, $views_front_arr); // not need default value !!!
        $this->addParam("Default_item_ID", "integer", 0, true);
        $this->addParam("Admin_page_size", "integer", \adminConfig::$adminRecordsPerPage, true);
        $this->addParam("Admin_default_controller", "select", reset($controllers_admin_arr), true, $controllers_admin_arr);
        $this->addParam("Admin_default_view", "select", "", true, $views_admin_arr); // not need default value !!!
        $this->addParam("Default_module_meta_title", "string", "");
        $this->addParam("Default_module_meta_description", "string", "");
        $this->addParam("Default_module_meta_keywords", "string", "");
        $params_file = PATH_MODULES . $this->getName() . DS . 'params.php';
        if (is_file($params_file)) {
            $className = $this->getName() . "ModuleParams";
            require_once $params_file;
            $className::_proceed($this);
        }
    }

    public function getParamsMask()
    {
        //if(!count($this->_paramsMask)) $this->setParamsMask();
        return $this->_paramsMask;
    }

    public function getACLTemplate($is_admin = true)
    {
        return array();
    }

    public function getLinksArray(&$counter, &$arr)
    {
        $counter++;
        $module = $this->getName();
        $arr[$module][$counter]['link'] = Router::_("index.php?module=" . $module, true);
        $arr[$module][$counter]['name'] = Text::_($module);
        $arr[$module][$counter]['fullname'] = Text::_($module);
        return true;
    }

    public function getSitemapHTML()
    {
        return array("html" => "", "title_link" => false, "links" => array());
    }

    public static function getInstance($name = '', $force_api = false)
    {
        if (!$name) $name = Registry::getInstance()->get('module');
        // name must be equal to [a-zA-Z0-9_]
        if (!preg_match('/^[\w]+$/', $name)) Util::redirect(Router::_("index.php"), Text::_("Page not found") . (defined("_ADMIN_MODE") || \siteConfig::$debugMode ? " => " . $name : ""), "404");
        // Check and load
        if (!array_key_exists($name, self::$_loadedModules)) {
            // Check in database and get it
            $db = Database::getInstance();
            $db->setQuery("SELECT * FROM #__modules WHERE m_name='" . $name . "' AND m_deleted=0");
            $moduleData = $db->loadObjectList();
            if (count($moduleData) != 1) {
                if ($name == \siteConfig::$defaultModule) self::fatalError(Text::_("Module not installed") . " => " . $name, defined("_ADMIN_MODE") ? "503" : "404"); // Realy fatal if default, else redirect already set in next lines.
                else Util::redirect(Router::_("index.php"), Text::_("Module not installed") . " => " . $name, "404");
            }
            if (!$moduleData[0]->m_enabled && !defined("_ADMIN_MODE") && !$force_api && self::$_loadMode == 'main' && !self::isCoreModule($name)) Util::redirect("/", Text::_("Module disabled") . " => " . $name . " (" . self::$_loadMode . ")", "404");
            // Module main script path
            $moduleScriptPath = PATH_MODULES . $name . DS . 'module.php';
            if (!file_exists($moduleScriptPath)) {
                if (
                    (defined("_ADMIN_MODE") && $name == \adminConfig::$adminDefaultModule)
                    ||
                    (!defined("_ADMIN_MODE") && $name == \siteConfig::$defaultModule)
                )
                    Util::fatalError(Text::_("Module not found") . " => " . $moduleScriptPath . ' => ' . $name, "404"); // Realy fatal if default, else redirect already set in next lines.
                else Util::redirect(Router::_("index.php"), Text::_("Module not found") . " => " . $moduleScriptPath . '=>' . $name, "404");
            }
            // Load language file
            if (!defined("_ADMIN_MODE")) Text::parseModule($name);
            Debugger::getInstance()->milestone("Including (require_once) " . $moduleScriptPath);
            require_once $moduleScriptPath;
            Debugger::getInstance()->milestone("Included (require_once) " . $moduleScriptPath);
            $moduleClass = $name . 'Module';
            self::$_loadedModules[$name] = new $moduleClass($name, ($force_api ? "api" : self::$_loadMode), $moduleData[0]);
            // For all next modules modes after `main`, mode must become `api`
            if (!$force_api) self::$_loadMode = 'api';
        }
        return self::$_loadedModules[$name];
    }

    //-----------------------------------------------------------
    public static function isCoreModule($moduleName)
    {
        return in_array($moduleName, self::$coreModules);
    }

    public static function isInstalled($moduleName)
    {
        $db = Database::getInstance();
        $db->setQuery("SELECT COUNT(m_id) FROM #__modules WHERE m_name='" . $moduleName . "'");
        if ($db->loadResult() != 1) return false;
        return true;
    }

    public static function getInstalledModules($forMap = false, $enabledOnly = false)
    {
        // Check in database
        $db = Database::getInstance();
        $sql = "SELECT DISTINCT m_name FROM #__modules WHERE m_deleted=0";
        if ($forMap) $sql .= " AND m_incl_map=1";
        if ($enabledOnly) $sql .= " AND m_enabled=1";
        $db->setQuery($sql);
        return $db->loadResultArray();
    }

    // private static $_config		= null;
    private $_config = array();
    private static $_helpers = array();

    public static function getHelper($helperName, $moduleName = "")
    {
        if (!$moduleName) $moduleName = Module::getInstance()->getName();
        $helperClass = $moduleName . 'Helper' . ucfirst($helperName);
        if (array_key_exists($helperClass, self::$_helpers) == false) {
            $helperScriptPath = PATH_MODULES . $moduleName . DS . 'helpers' . DS . $helperName . '.php';
            if (!file_exists($helperScriptPath)) {
                Util::fatalError(Text::_("Helper not found") . " => " . $moduleName . "." . $helperName . "(" . $helperScriptPath . ")", 404); // May be realy fatal. Method is static.
            }
            require_once $helperScriptPath;
            self::$_helpers[$helperClass] = new $helperClass();
        }

        return self::$_helpers[$helperClass];
    }

    public static function getControllers($moduleName = "", $front = true)
    {
        $files = array();
        if($moduleName) {
            $path = PATH_MODULES.$moduleName.DS."controllers".($front ? "" : DS."admin");
            $vfiles=Files::getFiles($path,false,false);
            foreach ($vfiles as $value) {
                $filename = basename($value["filename"],".php");
                $files[$filename]=$filename;
            }
        }
        return $files;
    }

    public static function getModules($front = true){
        $modules = array();
        $installed_modules = Module::getInstalledModules();
        $modules_dirs = Files::getFolders(PATH_MODULES,false,false);
        foreach($installed_modules as $_module) {
            // Check also count controllers
            if(count(Module::getControllers($_module, $front)) && array_key_exists($_module, $modules_dirs)) $modules[$_module] = $_module;
        }
        ksort($modules);
        return $modules;
    }

    //-----------------------------------------------------------

    protected $_controller = null;
    private $_controllerName = '';
    private $_defaultControllerName = 'default';
    private $_defaultViewName = 'default';
    private $_module_id = 0;
    private $_models = array();
    private $_showBreadcrumb = false;
    private $_disabled = '';
    protected $_adm_folder = "";

    public function __construct($name, $mode, $data)
    {
        Debugger::getInstance()->milestone("Module constructor start => " . $name);
        $this->initObj($name);
        $this->_adm_folder = defined("_ADMIN_MODE") ? "admin". DS : "";
        $_aclObject = ACLObject::getInstance($name . 'Module'); // canAccess() will be later
        if (!$_aclObject->canAccess()) { // DON'T move down and check only for "main", not for api. Both must be checked. Or it would be a hole.
            if ($name == \siteConfig::$defaultModule) $this->fatalError(Text::_('Permission is absent for') . " " . $_aclObject->getDescription()); // Realy fatal. Access to default module absent.
            else {
                if ((\siteConfig::$debugMode && User::getInstance()->isAdmin()) || \siteConfig::$debugMode > 100) $this->fatalError(Text::_('Permission is absent for') . " " . $_aclObject->getDescription(), defined("_ADMIN_MODE") ? "503" : "404"); // Render fatal for admin debug.
                else Util::redirect(Router::_("index.php"), Text::_('Permission is absent for') . " " . $_aclObject->getDescription(), "404");
            }
        } else {
            $config = $data->m_config;
            $this->_module_id = $data->m_id;
            $this->_showBreadcrumb = $data->m_show_breadcrumb;
            if ($mode == 'main') {
                $reestr = new FieldSet();
                $this->set('reestr', $reestr);
                $this->loadConfig($config);
                $this->prepare();
                $this->setDefaults();
                $this->initData();
                $this->getController();
            } else {
                $this->loadConfig($config);
            }
            $this->milestone("Module loaded => " . $name, __FUNCTION__);
        }
    }

    private function initData()
    {
        $controller = Request::getSafe('controller', $this->_defaultControllerName);
        $task = Request::getSafe('task', 'show');
        $view = Request::getSafe('view', $this->_defaultViewName);
        $layout = Request::getSafe('layout', 'default');

        $this->set('controller', $controller, true);
        $this->set('task', $task, true);
        $this->set('view', $view, true);
        $this->set('layout', $layout, true);
    }

    private function loadConfig($config)
    {
        $this->setParamsMask();
        $params = Params::parse($config);
        if (is_array($params)) $this->_config = $params; else $this->_config = array();
    }

    protected function setDefaults()
    {
        $default_controller = $this->getParam(defined("_ADMIN_MODE") ? "Admin_default_controller" : "Default_controller");
        if ($default_controller) $this->_defaultControllerName = $default_controller;

        $default_view = $this->getParam(defined("_ADMIN_MODE") ? "Admin_default_view" : "Default_view");
        if ($default_view) $this->_defaultViewName = $default_view;
    }

    public function getController()
    {
        // Load controller script
        $controllerName = $this->get('controller');
        if ($controllerName != $this->_controllerName) {
            $path = PATH_MODULES . $this->getName() . DS . 'controllers' . DS . $this->_adm_folder . $controllerName . '.php';
            if (!file_exists($path)) {
                if ($controllerName == "default") {
                    $this->fatalError(Text::_("Controller not found") . " => " . $this->getName() . "." . $controllerName); // Realy fatal if default, else redirect already set in next lines.
                } else {
                    if ((\siteConfig::$debugMode && User::getInstance()->isAdmin()) || \siteConfig::$debugMode > 100) $this->fatalError(Text::_("Controller not found") . " => " . $this->getName() . "." . $controllerName, defined("_ADMIN_MODE") ? "503" : "404"); // Render fatal for admin debug.
                    else Util::redirect(Router::_("index.php"), Text::_("Controller not found") . " => " . $this->getName() . "." . $controllerName, "404");
                }
            }
            Debugger::getInstance()->milestone("Including (require_once) " . $path);
            require_once $path;
            Debugger::getInstance()->milestone("Included (require_once) " . $path);
            $this->_controllerName = $controllerName;
            $controllerClass = $this->getName() . 'Controller' . ucfirst($controllerName);
            $this->_controller = new $controllerClass($controllerName, $this);
        }
        return $this->_controller;
    }

    public function getModel($modelName = "")
    {
        $moduleName = $this->getName();
        $viewName = $this->get('view');
        if (!$modelName) $modelName = $viewName;
        if (array_key_exists($modelName, $this->_models) == false) {
            $this->milestone("Model load attempt => MODULE=&quot;" . $moduleName . "&quot; MODEL=&quot;" . $modelName . "&quot; VIEW=&quot;" . $viewName . "&quot;", __FUNCTION__);
            $path = PATH_MODULES . $moduleName . DS . 'models' . DS . $this->_adm_folder . $modelName . '.php';
            if (!file_exists($path)) {
                if ($this->getDefaultView() == $viewName) {
                    $this->fatalError(Text::_("Model not found") . " => " . $moduleName . "." . $modelName); // Realy fatal if default, else redirect already set in next lines.
                } else {
                    // Don't remove this !!! Empty models are neccessary to prevent unpredictable behavior.
                    if ((\siteConfig::$debugMode && User::getInstance()->isAdmin()) || \siteConfig::$debugMode > 100) $this->fatalError(Text::_("Model not found") . " => " . $moduleName . "." . $modelName, defined("_ADMIN_MODE") ? "503" : "404"); // Render fatal for admin debug.
                    else Util::redirect(Router::_("index.php"), Text::_("Model not found") . " => " . $moduleName . "." . $modelName, "404");
                }
            }
            require_once $path;
            $modelClass = $moduleName . 'Model' . ucfirst($modelName);
            $this->_models[$modelName] = new $modelClass($this);
        }
        return $this->_models[$modelName];
    }

    public function getConfig()
    {
        return $this->_config;
    }

    public function getID()
    {
        return $this->_module_id;
    }

    public function getParam($param_name, $default_fill_behavior = true, $custom_default_value = null)
    {
         // Util::pre("---------------------------");
         // Util::pre($param_name, $default_fill_behavior, $custom_default_value);
        if(isset($this->_paramsMask[$param_name])){
            // Util::pre("Есть ключ в _paramsMask (1)");
            if(array_key_exists($param_name, $this->_config)){
                // Util::pre("Есть ключ в this->_config (2)");
                switch ($this->_paramsMask[$param_name]["vtype"]) {
                    case "boolean": // тут возвращаем сразу
                        if ($this->_config[$param_name]) return 1;
                        else return 0;
                        break;
                    case "integer":
                        $val = intval($this->_config[$param_name]);
                        break;
                    case "float":
                        $val = floatval($this->_config[$param_name]);
                        break;
                    case "date":
                        $val = $this->_config[$param_name];
                        break;
                    case "table_select":
                    case "select":
                    case "string":
                    case "text":
                    case "ro_string":
                    case "title":
                    default: // это все считаем строками
                        $val = $this->_config[$param_name];
                        break;
                }
                // Util::pre("Значение из конфига = ".$val." (3)");
                if (!$val) {
                    // Util::pre("Значение из конфига пустое (4)");
                    if(!$default_fill_behavior /* && !is_null($custom_default_value) */){
                        // Util::pre("Оверрайд заполнения по дефолту (5)");
                        // Return even null value
                        return $custom_default_value;
                    } else {
                        // Util::pre("Работаем по дефолтам _paramsMask (6)");
                        // Util::pre("this->_paramsMask[param_name][fill_default] = ".$this->_paramsMask[$param_name]["fill_default"]);
                        if(isset($this->_paramsMask[$param_name]["fill_default"])
                            && $this->_paramsMask[$param_name]["fill_default"]
                            && isset($this->_paramsMask[$param_name]["vdefault"]))
                        {
                            return $this->_paramsMask[$param_name]["vdefault"];
                        } else {
                            return null;
                        }
                    }
                } else {
                    return $val;
                }
            } else {
                // Util::pre("Нет ключа в this->_config (11)");
                // Param absent in config
                if(!$default_fill_behavior /* && !is_null($custom_default_value) */){
                    // Util::pre("Оверрайд заполнения по дефолту (12)");
                    // Return even null value
                    return $custom_default_value;
                } else {
                    // Util::pre("Работаем по дефолтам _paramsMask (13)");
                    // Util::pre("this->_paramsMask[param_name][fill_default] = ".$this->_paramsMask[$param_name]["fill_default"]);
                    if(isset($this->_paramsMask[$param_name]["fill_default"])
                        && $this->_paramsMask[$param_name]["fill_default"]
                        && isset($this->_paramsMask[$param_name]["vdefault"]))
                    {
                        return $this->_paramsMask[$param_name]["vdefault"];
                    } else {
                        return null;
                    }
                }
            }
        } else {
            // Util::pre("Нет ключа в _paramsMask (21)");
            // Param absent in params
            if(!$default_fill_behavior /* && !is_null($custom_default_value) */){
                // Util::pre("Оверрайд заполнения по дефолту (22)");
                // Return even null value
                return $custom_default_value;
            } else {
                // Util::pre("Нет оверрайда (23)");
                return null;
            }
        }

        /*
        // @FIXME DELETE LATER
        if (array_key_exists($param_name, $this->_config) && isset($this->_paramsMask[$param_name])) {
            switch ($this->_paramsMask[$param_name]["vtype"]) {
                case "boolean": // тут возвращаем сразу
                    if ($this->_config[$param_name]) return 1;
                    else return 0;
                    break;
                case "integer":
                    $val = intval($this->_config[$param_name]);
                    break;
                case "float":
                    $val = floatval($this->_config[$param_name]);
                    break;
                case "table_select":
                case "select":
                case "string":
                case "text":
                case "ro_string":
                case "title":
                default: // это все считаем строками
                    $val = $this->_config[$param_name];
                    break;
            }
            if (!$val && isset($this->_paramsMask[$param_name]["fill_default"]) && $this->_paramsMask[$param_name]["fill_default"]) {
                if (!is_null($custom_default_value)) return $custom_default_value;
                elseif ($default_fill_behavior && isset($this->_paramsMask[$param_name]["vdefault"])) return $this->_paramsMask[$param_name]["vdefault"];
            }
            return $val;
        } else {
            if ($default_fill_behavior && isset($this->_paramsMask[$param_name]["vdefault"])) return $this->_paramsMask[$param_name]["vdefault"];
            elseif (!is_null($custom_default_value)) return $custom_default_value;
            else return "";
        }
        */
    }

    public function setBreadCrumbVisibility($status = 1)
    {
        $this->_showBreadcrumb = $status;
    }

    public function showBreadcrumb()
    {
        return $this->_showBreadcrumb;
    }

    //-------------------- Interface methods --------------------
    public function prepare()
    {
        // Just for overriding
    }

    public function render()
    {
        // Javascript for module
        if (!Portal::getInstance()->isDisabled() && !Portal::getInstance()->inPrintMode()) {
            Portal::getInstance()->addScript("modules/" . $this->getName() . ".js");
        }
        if ($this->_disabled) echo $this->_disabled;
        else $this->_controller->executeTask();
    }

    public function ajax()
    {
        if ($this->_disabled) echo $this->_disabled;
        else $this->_controller->executeAjax();
    }

    public function getBreadCrumbArray()
    {
        return $this->_controller->getView()->getBreadCrumbArray();
    }

    public function renderBreadCrumb($force = false)
    {
        if ($this->showBreadcrumb() || $force) return $this->_controller->getView()->renderBreadCrumb();
        else return "";
    }

    public function setDefaultView($viewName)
    {
        $this->_defaultViewName = $viewName;
    }

    public function getDefaultView()
    {
        return $this->_defaultViewName;
    }

    public function _disable($msg = "Module disabled")
    {
        if (!$this->_disabled) $this->_disabled = "<div id=\"disabled-message\">" . $msg . "</div>";
    }
}

?>